import logging
import os
from logging.config import dictConfig

from flask import has_request_context, request


class RequestFormatter(logging.Formatter):
    def format(self, record):
        record.request_id = 'NA'

        if has_request_context():
            record.request_id = request.environ.get("HTTP_X_REQUEST_ID")

        return super().format(record)


def setup_logging(debug_file, error_file):
    os.makedirs(os.path.dirname(debug_file), exist_ok=True)
    os.makedirs(os.path.dirname(error_file), exist_ok=True)
    dictConfig({
        'version': 1,
        'formatters': {'default': {
            '()': 'CYLGame.Log.RequestFormatter',
            'format': '[%(asctime)s] %(levelname)s in %(module)s for %(request_id)s : %(message)s',
        }, 'error': {
            '()': 'CYLGame.Log.RequestFormatter',
            'format': '[%(asctime)s] %(levelname)s in %(module)s for %(request_id)s : %(message)s',
        }},
        'handlers': {'wsgi': {
            'class': 'logging.StreamHandler',
            'level': 'WARN',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'default'
        }, 'debug': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'level': 'DEBUG',
            'filename': debug_file,
            'backupCount': 2,
            'formatter': 'default'
        }, 'error': {
            'class': 'logging.handlers.RotatingFileHandler',
            'level': 'ERROR',
            'filename': error_file,
            'maxBytes': 10000000,
            'backupCount': 1,
            'formatter': 'error'
        }},
        'root': {
            'level': 'DEBUG',
            'handlers': ['wsgi', 'debug', 'error']
        }
    })


from __future__ import print_function
from __future__ import division

import sys
from itertools import islice
from random import randint, choice, shuffle

from multiprocessing import Process, Event, Pool

import time

import math
from typing import Dict

from CYLGame.Utils import OnlineMean, choose
from .Game import GameRunner
from .Player import Room


def create_room(gamedb, bot, compiler, size):
    stokens = gamedb.get_school_tokens()
    pool = []
    for stoken in stokens:
        pool += gamedb.get_tokens_for_school(stoken)

    if len(pool) == 0:
        from copy import deepcopy
        return Room([bot] + [deepcopy(bot) for _ in range(size - 1)])

    bots = [bot]
    while True:
        token = choice(pool)
        code, options = gamedb.get_code_and_options(token)
        try:
            prog = compiler.compile(code)
            prog.options = options
            bots += [prog]
            if len(bots) >= size:
                return Room(bots)
        except:
            print("Couldn't compile code for '{}' in '{}'".format(token, gamedb.get_school_for_token(token)))


def avg(scores):
    return float((sum(scores) * 100) / len(scores)) / 100


def sim_prog_for_score(game, compiler, code, options, seed, debug=True):
    runner = GameRunner(game)
    prog = compiler.compile(code)
    prog.options = options
    try:
        score = runner.run(Room([prog], seed=seed), playback=False).score
        sys.stdout.write(".")
        sys.stdout.flush()
        return score
    except:
        print("There was an error simulating the game. Make sure this isn't a bug with the program.\n  Seed: {}\n  Code: '''{}'''".format(seed, code))
        return 0


# TODO: rewrite this function. It is very outdated!
def sim_competition(compiler, game, gamedb, token, runs, ncores=None, debug=False, score_func=avg):
    """

    Args:
        compiler:
        game:
        gamedb:
        token:
        runs:
        ncores:      If None will default to the number of all available cores.
        debug:
        score_func:

    Returns:

    """
    assert gamedb is not None
    assert gamedb.is_comp_token(token)

    seeds = [randint(0, sys.maxsize) for _ in range(runs)]

    for school in gamedb.get_schools_in_comp(token):
        if debug:
            print("Got school '" + school + "'")
        max_score = 0
        max_code = ""
        for student in gamedb.get_tokens_for_school(school):
            if debug:
                print("Got student '" + student + "'")
            code, options = gamedb.get_code_and_options(student)
            if not code:
                continue
            if debug:
                print("Compiling code...")
            prog = compiler.compile(code)
            prog.options = options
            if debug:
                print("Simulating {} games...".format(runs))

#             # TODO: make this able to run in a pool of threads (so it can be run on multiple CPUs)
            with Pool(processes=ncores) as pool:
                scores = pool.starmap(sim_prog_for_score, [(game, compiler, code, options, seed, debug) for seed in seeds])

            score = score_func(scores)
            if debug:
                print(" Score: {}".format(score))

            if score > max_score:
                max_score = score
                max_code = code
        if debug:
            print("Saving score...", max_score)
        gamedb.set_comp_avg_score(token, school, max_score)
        gamedb.set_comp_school_code(token, school, max_code)
    if debug:
        print("All done :)")


class Ranking(object):
    def __init__(self, bots):
        """

        Args:
            bots: Tuple containing standing and bot id
        """
        self.ranks = {}
        for i in range(0, len(bots)):
            self.ranks[bots[i]] = i  # Key: BOTS Value: RANK

    def __add__(self, other):
        new_rank = {}
        for k in self.ranks:
            new_rank[k] = self.ranks[k] + other[k]
        return new_rank

    def __radd__(self, other):
        if other == 0:
            return self
        new_rank = {}
        for k in self.ranks:
            new_rank[k] = self.ranks[k] + other[k]
        return new_rank

    def add_rank(self, standing, bot):
        self.ranks[bot] += standing


class MultiplayerComp(object):
    def __init__(self, bots, room_size, default_bot_class, run_factor=10):
        """

        Args:
            bots: (PROG) the players program to be executed and ranked
            room_size: (INT) the size of the rooms
            run_factor(int): The average number of random games per combination of bots.
        """
        self.default_bot_class = default_bot_class
        self.room_size = room_size
        self.scores = {}  # Prog:scores
        self.rooms = {}  # Rooms:Rankings
        self.cur_run = 0
        self.total_runs = run_factor * choose(max(len(bots), 4), 4)
        for bot in bots:
            self.scores[bot] = OnlineMean()

    def __iter__(self):
        return self

    def __setitem__(self, key, value):
        """
        ARGS:
            key (Room): The room to set to ranking
            value (Ranking): The returned ranking
        """
        self.rooms[key] = value
        for k in value.ranks:
            if k.prog in self.scores:
                self.scores[k.prog] += value.ranks[k] * 10

    def __getitem__(self, item):
        return self.scores[item]

    def __next__(self):
        return self.next()

    def next(self):
        if self.cur_run == self.total_runs:
            raise StopIteration()

        l = list(self.scores.keys())
        shuffle(l)
        p = l[:self.room_size + 1]
        while len(p) < self.room_size:
            p += [self.default_bot_class()]
        room = Room(p)
        self.rooms[room] = None
        self.cur_run += 1
        return room

    @staticmethod
    def sim_multiplayer(s_token, gamedb, game, compiler, save_games=False, debug=False):
        assert game.MULTIPLAYER, "Game must be multi-player to do MultiplayerComp."
        assert gamedb is not None
        assert gamedb.is_school_token(s_token)
        students = gamedb.get_tokens_for_school(s_token)  # Only getting one school token
        bots = []
        game_tokens = []
        for s in students:
            try:
                if debug:
                    print("got student '" + s + "'")
                code, options = gamedb.get_code_and_options(s)
                name = gamedb.get_name(s)
                if not code:
                    continue
                if debug:
                    print("compiling code...")
                prog = compiler.compile(code)
                prog.options = options
                prog.token = s
                prog.name = name
                if debug:
                    print("simulating...")
                bots += [prog]
            except:
                print("Couldn't compile code for '{}' in '{}'".format(s, gamedb.get_school_for_token(s)))
        if len(bots) == 0:
            if debug:
                print("School Token {} has no valid bots. :(".format(s_token))
            return []
        tourney = MultiplayerComp(bots, game.get_number_of_players(), game.default_prog_for_computer())
        for room in tourney:
            if debug:
                print("Room: " + str(room))
            gamerunner = GameRunner(game)
            tourney[room] = gamerunner.run(room).score
            if save_games:
                game_tokens += [room.save(gamedb)]

        for player, score in tourney.scores.items():
            gamedb.save_avg_score(player.token, score.floored_mean)
            if debug:
                print("Score {} for Bike: {}".format(score.mean, str(player.token)))

        return game_tokens

    @staticmethod
    def make_bot(gamedb, compiler, token, debug=False):
        try:
            if debug:
                print("Making bot for '" + token + "'")
            code, options = gamedb.get_code_and_options(token)
            name = gamedb.get_name(token)
            if not code:
                return
            if debug:
                print("compiling code...")
            prog = compiler.compile(code)
            prog.options = options
            prog.token = token
            prog.name = name
            return prog
        except:
            print("Couldn't compile code for '{}' in '{}'".format(s, gamedb.get_school_for_token(s)))

    @staticmethod
    def sim_comp(c_token, gamedb, game, compiler, save_games=False, debug=False):
        if debug:
            print("Getting the best bots from each school...")

        # Get the best valid bot
        best_bots = {}
        for s_token in gamedb.get_schools_in_comp(c_token):
            best_bot = None
            best_bot_score = float("-inf")
            for token in gamedb.get_tokens_for_school(s_token):
                if gamedb.get_avg_score(token, float("-inf")) > best_bot_score:
                    bot = MultiplayerComp.make_bot(gamedb, compiler, token, debug=debug)

                    if bot:
                        # Was compilable and the best bot
                        best_bot_score = gamedb.get_avg_score(token)
                        best_bot = bot
            if best_bot:
                best_bots[s_token] = best_bot
            elif debug:
                print(f"School {s_token} had on bot with a score")

        if debug:
            print(f"Got the following bots:\n{best_bots}")

        if len(best_bots) == 0:
            if debug:
                print("No School had any valid bots. :(")
            return

        game_tokens = []
        tourney = MultiplayerComp(best_bots.values(), game.get_number_of_players(), game.default_prog_for_computer())
        if debug:
            print(f"Will run {tourney.total_runs:.0f} games.")
        start_time = time.time()
        for i, room in enumerate(tourney):
            if debug:
                time_taken = time.time()-start_time
                estimate_total_time = time_taken * (tourney.total_runs/(i+1))
                eta = estimate_total_time - time_taken
                print(f"Eta: ~{eta:.0f} secs Cur Room: {room}")
            gamerunner = GameRunner(game)
            tourney[room] = gamerunner.run(room).score
            if save_games:
                game_tokens += [room.save(gamedb)]

        if save_games:
            gamedb.replace_games_in_comp(c_token, new_gtokens=save_games)

        # Save scores
        for s_token, best_bot in best_bots.items():
            best_bot_score = tourney[best_bot].rounded_mean
            gamedb.set_comp_avg_score(c_token, s_token, best_bot_score)


class MultiplayerCompRunner(Process):
    def __init__(self, interval, gamedb, game, compiler, debug=False):
        """

        Args:
            interval(int): The number of seconds between simulating. This is the time between the start of one run and
                the start of the next. Therefore a interval of 0 means when done running once start again.
            gamedb:
            game:
            compiler:
        """
        super(MultiplayerCompRunner, self).__init__()
        self.interval = interval
        self.gamedb = gamedb
        self.game = game
        self.compiler = compiler
        self.start_run = Event()
        self.end = Event()
        self.debug = debug

    def stop(self):
        self.start_run.set()
        self.end.set()

    def run(self):
        while not self.end.is_set():
            if not self.start_run.is_set():
                self.__run()
            self.start_run.wait(self.interval)

    def __run(self):
        print("Scoring all Bots...")
        start_time = time.time()
        gtokens = []
        for s_token in self.gamedb.get_school_tokens():
            gtokens += MultiplayerComp.sim_multiplayer(s_token, self.gamedb, self.game, self.compiler,
                                                       save_games=True, debug=self.debug)
        self.gamedb.replace_games_in_comp(ctoken="P00000000",  # TODO: un hardcode this.
                                          new_gtokens=gtokens)
        print("Finished scoring in {0:.2f} secs...".format(time.time() - start_time))


class RollingMultiplayerComp(object):
    def __init__(self, room_size, default_bot_class, rolling_n):
        """

        Args:
            room_size: (INT) the size of the rooms
            rolling_n: (int) the number of games each players score should be averaged from.
        """
        self.default_bot_class = default_bot_class
        self.room_size = room_size
        self.scores = {}  # type: Dict[Prog, OnlineMean]
        self.rooms = {}  # Rooms:Rankings
        self.rolling_n = rolling_n

        self.bots = set()  # A set of tokens of the already added bots.

    def add_bot(self, bot, gamedb):
        # load old mean
        mean = gamedb.get_value(bot.token, "rolling_score", 0)
        n = gamedb.get_value(bot.token, "rolling_n", 0)
        self.scores[bot] = OnlineMean(n, mean, roll_after_n=self.rolling_n)
        self.bots |= {bot.token}

    def save_rolling_scores(self, gamedb):
        for bot, score in self.scores.items():
            gamedb.save_value(bot.token, "rolling_score", score.mean)
            gamedb.save_value(bot.token, "rolling_n", score.i)
            gamedb.save_avg_score(bot.token, score.floored_mean)

    def add_bot_if_needed(self, bot, gamedb):
        if bot.token not in self.bots:
            self.add_bot(bot, gamedb)

    def __iter__(self):
        return self

    def __setitem__(self, key, value):
        """
        ARGS:
            key (Room): The room to set to ranking
            value (Ranking): The returned ranking
        """
        self.rooms[key] = value
        for k in value.ranks:
            if k.prog in self.scores:
                self.scores[k.prog] += value.ranks[k] * 10

    def __next__(self):
        return self.next()

    def next(self):
        l = list(self.scores.keys())
        shuffle(l)
        p = l[:self.room_size + 1]
        while len(p) < self.room_size:
            p += [self.default_bot_class()]
        room = Room(p)
        self.rooms[room] = None
        return room


class RollingMultiplayerCompRunner(Process):
    def __init__(self, interval, gamedb, game, compiler, rolling_n=100, batch_size=4, debug=False):
        """

        Args:
            interval(int): The number of seconds between simulating each game. Should be a pretty small number.
            gamedb:
            game:
            compiler:
        """
        super(RollingMultiplayerCompRunner, self).__init__()

        assert game.MULTIPLAYER, "Game must be multi-player to do MultiplayerComp."
        assert gamedb is not None

        self.interval = interval
        self.gamedb = gamedb
        self.game = game
        self.compiler = compiler
        self.rolling_n = rolling_n
        self.batch_size = batch_size
        self.debug = debug

        self.start_run = Event()
        self.end = Event()

        self.comps = {}  # type: Dict[str, RollingMultiplayerComp]
        self.i = -1  # This is to make the clean happen on the first run. This way it will run at least every time you start the server.

    def stop(self):
        self.start_run.set()
        self.end.set()

    def run(self):
        while not self.end.is_set():
            if not self.start_run.is_set():
                self.__run()
            self.start_run.wait(self.interval)

    def __run(self):
        print("Scoring a couple games...")
        self.i += 1
        start_time = time.time()
        # gtokens = []
        for s_token in self.gamedb.get_school_tokens():
            if s_token not in self.comps:
                self.comps[s_token] = RollingMultiplayerComp(room_size=self.game.get_number_of_players(),
                                                             default_bot_class=self.game.default_prog_for_computer(),
                                                             rolling_n=self.rolling_n)
        for s_token, comp in self.comps.items():
            # First make sure all the bots in the school are in the comp.
            student_tokens = self.gamedb.get_tokens_for_school(s_token)
            for token in student_tokens:
                # only add bot if not already added.
                if token not in comp.bots:
                    bot = self.make_bot(token)
                    if bot:
                        comp.add_bot(bot, self.gamedb)

            if len(comp.bots) == 0:
                if self.debug:
                    print("School Token {} has no valid bots. :(".format(s_token))
                continue  # skip this school if it has no bots to run.

            # Now let's run a couple games.
            for room in islice(comp, self.batch_size):
                if self.debug:
                    print("Room: " + str(room))
                gamerunner = GameRunner(self.game)
                comp[room] = gamerunner.run(room).score

                # save scores and game
                self.gamedb.add_game_to_comp("P00000000", room.save(self.gamedb))
                comp.save_rolling_scores(self.gamedb)

        # TODO: replace this with a function which keeps the last rolling_n number of games for each token.
        # self.gamedb.replace_games_in_comp(ctoken="P00000000",  # TODO: un hardcode this.
        #                                   new_gtokens=gtokens)
        print("Finished scoring games in {0:.2f} secs...".format(time.time() - start_time))
        if self.i % self.rolling_n == 0:
            start_time = time.time()
            print("Will do some clean up....")
            self.clean_up_old_games()
            print("Finished cleaning in {0:.2f} secs...".format(time.time() - start_time))

    def make_bot(self, token):
        try:
            if self.debug:
                print("Making bot for '" + token + "'")
            code, options = self.gamedb.get_code_and_options(token)
            name = self.gamedb.get_name(token)
            if not code:
                return
            if self.debug:
                print("compiling code...")
            prog = self.compiler.compile(code)
            prog.options = options
            prog.token = token
            prog.name = name
            return prog
        except:
            print("Couldn't compile code for '{}' in '{}'".format(s, gamedb.get_school_for_token(s)))

    def clean_up_old_games(self):
        # Since we keep making new games we should clean up the old ones.
        # Only keep the newest `rolling_n` number of games for each player
        all_game_tokens = set(self.gamedb.get_games_for_token("P00000000"))
        games_to_delete = set(all_game_tokens)

        for s_token in self.gamedb.get_school_tokens():
            for token in self.gamedb.get_tokens_for_school(s_token):
                games = set(self.gamedb.get_games_for_token(token))
                games = games.intersection(all_game_tokens)  # only keep the games that were used for scoring.
                if self.rolling_n >= len(games):  # We need to keep them all
                    games_to_delete = games_to_delete - games
                else:  # This player has too many games.
                    # We need to pick the newest ones
                    games = list(games)
                    games.sort(key=lambda x: self.gamedb.get_ctime_for_game(x), reverse=True)
                    games_player_needs = games[:self.rolling_n]
                    games_to_delete = games_to_delete - set(games_player_needs)

        if self.debug:
            print("Cleaning up {} games...".format(len(games_to_delete)))

        for gtoken in games_to_delete:
            self.gamedb.delete_game(gtoken)

    # def clean_up_broken_games(self):
    #     all_recorded_games = set(self.gamedb.get_all_game_tokens())
    #     for s_token in self.gamedb.get_school_tokens():
    #         for token in self.gamedb.get_tokens_for_school(s_token):
    #             for game_token in self.gamedb.get_games_for_token(token):
    #                 if game_token not in all_recorded_games:
    #                     print("This is bad!!!")
    #     for comp_token in self.gamedb.get_all_comp_tokens():
    #         for game_token in self.gamedb.get_games_for_token(token):
    #             if game_token not in all_recorded_games:
    #                 print("This is bad!!!")




from __future__ import division
import os.path
import random
import sys
from dataclasses import dataclass
from typing import List, Optional

from CYLGame.Frame import GridFrameBuffer
from CYLGame.Player import UserProg, Player
from CYLGame.Utils import int2base

FPS = 30


def scorer(func):
    """This function is a decorator for a scoring function.
       This is hack a to get around self being passed as the first argument to the scoring function."""
    def wrapped(a, b=None):
        if b is not None:
            return func(b)
        return func(a)
    return wrapped


@scorer
def average(scores):
    return float((sum(scores) * 100) / len(scores)) / 100


def data_file(filename):
    resource_path = os.path.join(os.path.split(__file__)[0], "data", filename)
    return resource_path


class GameLanguage(object):
    LITTLEPY = 0

    @staticmethod
    def get_language_description(language):
        if language == GameLanguage.LITTLEPY:
            return open(data_file("little_python_intro.md")).read()


class Game(object):
    WEBONLY = True
    SCREEN_WIDTH = 0
    SCREEN_HEIGHT = 0
    GAME_TITLE = ""
    OPTIONS = None
    MULTIPLAYER = False  # TODO: document

    def is_running(self):
        """This is how the game runner knows if the game is over.

        Returns:
            bool: True if the game should still be play. False otherwise
        """
        raise Exception("Not implemented!")

    def create_new_player(self, prog):
        """This creates n new objects that inherits from the Player class.

        Returns:
            n new objects that inherit from the Player class with the given program.
        """
        raise Exception("Not implemented!")

    def get_debug_vars(self):
        raise Exception("Not implemented!")

    def do_turn(self):
        raise Exception("Not implemented!")

    def get_frame(self):
        raise Exception("Not implemented!")

    def init_board(self):
        raise Exception("Not implemented!")

    def start_game(self):
        raise Exception("Not implemented!")

    @staticmethod
    def default_prog_for_bot(language):
        raise Exception("Not implemented!")

    @staticmethod
    def default_prog_for_computer():
        """This method is only for multi-player games to implement.
        Returns:
            type[Prog]
        """
        raise Exception("Not implemented!")

    @staticmethod
    def get_intro():
        raise Exception("Not implemented!")

    @staticmethod
    def get_move_consts():
        return ConstMapping()

    @staticmethod
    def get_number_of_players():
        """This method is only for multi-player games to implement.
        Returns:
            int: The number of players needed to play the game.
        """
        raise Exception("Not implemented!")


class NonGridGame(Game):
    WEBONLY = True
    GRID = False

    def read_bot_state(self, state):
        raise Exception("Not Implemented!")

    def get_vars_for_bot(self):
        raise Exception("Not Implemented!")


class ConstMapping:
    def __init__(self, seq=None, **kwargs):
        """
        dict() -> new empty dictionary
        dict(mapping) -> new dictionary initialized from a mapping object's
            (key, value) pairs
        dict(iterable) -> new dictionary initialized as if via:
            d = {}
            for k, v in iterable:
                d[k] = v
        dict(**kwargs) -> new dictionary initialized with the name=value pairs
            in the keyword argument list.  For example:  dict(one=1, two=2)
        # (copied from class doc)
        """
        self.name_to_val_mapping = {}
        self.val_to_name_mapping = {}
        if isinstance(seq, dict):
            for v, k in seq.items():
                self[v] = k
        elif kwargs:
            for v, k in kwargs:
                self[v] = k
        else:
            for v, k in seq:
                self[v] = k

    def update(self, other):
        for var, val in other.items():
            self[var] = val

    @property
    def names(self):
        return self.name_to_val_mapping.keys()

    @property
    def values(self):
        return self.name_to_val_mapping.values()

    def __iter__(self):
        return iter(self.name_to_val_mapping.items())

    def __contains__(self, item):
        return item in self.name_to_val_mapping or item in self.val_to_name_mapping

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.name_to_val_mapping[item]
        elif isinstance(item, int):
            return self.val_to_name_mapping[item]
        raise TypeError()

    def __setitem__(self, key, value):
        # TODO: assert key is a valid littlepython variable name.
        assert isinstance(key, str), "Key must be a variable name"
        assert isinstance(value, int), "Only valid value is an int currently"

        if key in self.name_to_val_mapping:
            old_value = self.name_to_val_mapping[key]
            del self.name_to_val_mapping[key]
            del self.val_to_name_mapping[old_value]
        assert value not in self.val_to_name_mapping, "This is a one-to-one mapping"

        self.name_to_val_mapping[key] = value
        self.val_to_name_mapping[value] = key

    def __delitem__(self, key):
        value = self.name_to_val_mapping[key]
        del self.name_to_val_mapping[key]
        del self.val_to_name_mapping[value]

    def __len__(self):
        return len(self.name_to_val_mapping)


class GridGame(Game):
    WEBONLY = False
    GRID = True
    SCREEN_WIDTH = 80
    SCREEN_HEIGHT = 25
    CHAR_WIDTH = 8
    CHAR_HEIGHT = 8
    GAME_TITLE = ""
    CHAR_SET = data_file("fonts/terminal8x8_gs_ro.png")
    __frame_buffer = None

    def is_running(self):
        """This is how the game runner knows if the game is over.

        Returns:
            bool: True if the game should still be play. False otherwise
        """
        raise Exception("Not implemented!")

    def create_new_player(self, prog):  # TODO: add option to create computer, user or bot players.
        """ TODO: write this
        """
        raise Exception("Not Implemented!")

    def do_turn(self):
        """This function should read the new state of all the players and react to them."""
        raise Exception("Not implemented!")

    def draw_screen(self, frame_buffer):
        """WARNING: There MUST NOT be any game logic in this function since it isn't called when simulating the game
                    during the competitions.
        """
        raise Exception("Not implemented!")

    def get_frame(self):
        if self.__frame_buffer is None:
            self.__frame_buffer = GridFrameBuffer(self.SCREEN_WIDTH, self.SCREEN_HEIGHT)
        self.draw_screen(self.__frame_buffer)
        return self.__frame_buffer.dump()

    def get_vars(self, player):
        """ TODO: write this
        """
        raise Exception("Not implemented!")

    def get_score(self):
        """This is the game runner gets the ending or mid-game score

        Returns:
            int: The current score.
        """
        raise Exception("Not implemented!")

    @staticmethod
    def default_prog_for_bot(language):
        raise Exception("Not implemented!")

    @staticmethod
    def get_intro():
        raise Exception("Not implemented!")

    @staticmethod
    def get_move_consts():
        return ConstMapping({"north": ord("w"), "south": ord("s"), "west": ord("a"), "east": ord("d"),
                "northeast": ord("e"), "southeast": ord("c"), "northwest": ord("q"), "southwest": ord("z")})


@dataclass
class PlayGameState:
    game: Game
    computer_players: List[Player]
    human_player: Player
    seed: int

    moves: str
    frame: Optional[List[List[int]]]


class GameRunner(object):
    def __init__(self, game_class):
        self.game_class = game_class  # type: Type[Game]

    def run(self, room, playback=True):
        """The the game.

        Args:
            room(Room): The room to play.
            playback(bool): if True save the playback.

        Returns:
            Room:
        """
        assert len(room.bots) > 0  # Make sure that we have a bot to run

        game = self.game_class(random.Random(room.seed))

        game.init_board()
        players = []
        for bot in room.bots:
            # TODO: This is a hack. Remove me sometime.
            if not hasattr(bot, "options"):
                bot.options = {}
            if playback:
                bot.options["debug"] = True

            players += [game.create_new_player(bot)]

        game.start_game()

        screen_cap = []
        while game.is_running():
            if playback:
                screen_cap += [game.get_frame()]

            for player in players:
                player.run_turn(game.random)

            game.do_turn()

        room.score = game.get_score()
        if playback:
            room.screen_cap = screen_cap
            for player in players:
                room.debug_vars[player.prog] = player.debug_vars

        return room

    def run_for_avg_score(self, room, times=1, func=average):
        """Runs the given game keeping only the scores.

        Args:
            times (int): The number of times to run to get the average score.

        Return:
            The return value the average score for the times runs.
        """
        # TODO: replace this method with a better one.
        scores = []
        for t in range(times):
            scores += [self.run(room.rand_seeded).score]
        return func(scores)

    def run_with_local_display(self, seed=None):
        """Will run the game for a user.

        Returns:
        """
        # This import statement is here so pygame is only imported if needed.
        from CYLGame import Display
        game = self.game_class(random.Random(seed))

        charset = Display.CharSet(self.game_class.CHAR_SET, self.game_class.CHAR_WIDTH, self.game_class.CHAR_HEIGHT)
        display = Display.PyGameDisplay(*charset.char_size_to_pix((game.SCREEN_WIDTH, game.SCREEN_HEIGHT)), title=game.GAME_TITLE)

        clock = Display.get_clock()

        frame_buffer = GridFrameBuffer(*charset.pix_size_to_char(display.get_size()), charset=charset)
        frame_updated = True

        game.init_board()
        players = []
        if game.MULTIPLAYER:
            computer_bot_class = game.default_prog_for_computer()
            for _ in range(game.get_number_of_players() - 1):
                players += [game.create_new_player(computer_bot_class())]
        player = game.create_new_player(UserProg())

        game.start_game()

        while game.is_running():
            clock.tick(FPS)
            for key in display.get_keys():
                # TODO: fix
                player.prog.key = key
                player.run_turn(game.random)
                for comp_player in players:
                    comp_player.run_turn(game.random)
                game.do_turn()
                frame_updated = True

            if frame_updated:
                game.draw_screen(frame_buffer)
                display.update(frame_buffer)
                frame_updated = False

    def init_game(self, seed: Optional[int]=None) -> PlayGameState:
        game = self.game_class(random.Random(seed))

        game.init_board()
        players = []
        if game.MULTIPLAYER:
            computer_bot_class = game.default_prog_for_computer()
            for _ in range(game.get_number_of_players() - 1):
                players += [game.create_new_player(computer_bot_class())]
        player = game.create_new_player(UserProg())

        game.start_game()
        return PlayGameState(game=game, computer_players=players, human_player=player, seed=seed, moves='', frame=game.get_frame())

    @staticmethod
    def move_game(state: PlayGameState, move: str) -> PlayGameState:
        player, comp_players = state.human_player, state.computer_players
        local_random = state.game.random
        game = state.game

        player.prog.key = move
        player.run_turn(local_random)
        for comp_player in state.computer_players:
            comp_player.run_turn(local_random)
        game.do_turn()
        state.frame = game.get_frame()

        if game.is_running():
            state.moves += move

        return state


def run(game_class, avg_game_func=average):
    def serve(args):
        print("I am going to serve")
        from .Server import serve
        reuse_addr = True
        if args.no_addr_reuse:
            reuse_addr = None
        serve(game_class, host=args.host, port=args.port, game_data_path=args.dbfile, avg_game_func=avg_game_func,
              debug=args.debug, multiplayer_scoring_interval=args.scoring_time, reuse_addr=reuse_addr)

    def play(args):
        print("Playing...")
        GameRunner(game_class).run_with_local_display(int(args.seed, 36))

    import argparse

    parser = argparse.ArgumentParser(prog=game_class.GAME_TITLE, description='Play ' + game_class.GAME_TITLE + '.')
    subparsers = parser.add_subparsers(help='What do you what to do?')
    subparsers.required = True
    subparsers.dest = "command"
    parser_play = subparsers.add_parser('play', help='Play ' + game_class.GAME_TITLE + ' with a GUI')
    parser_play.add_argument('-s', '--seed', nargs="?", type=str, help='Manually set the random seed.',
                             default=int2base(random.randint(0, sys.maxsize), 36))
    parser_play.set_defaults(func=play)
    parser_serve = subparsers.add_parser('serve', help='Serve ' + game_class.GAME_TITLE + ' to the web.')
    parser_serve.add_argument('-p', '--port', nargs="?", type=int, help='Port to serve on', default=5000)
    parser_serve.add_argument('-db', '--dbfile', nargs="?", type=str, help='The root path of the game database', default="temp_game")
    parser_serve.add_argument('--debug-log', nargs="?", type=str, help='The file path to use for the debug log.'
                                                                       '"{dbname}" in the path will be replaced by the path to the game database', default="{dbfile}/log/debug.log")
    parser_serve.add_argument('--error-log', nargs="?", type=str, help='The file path to use for the error log.'
                                                                       '"{dbname}" in the path will be replaced by the path to the game database', default="{dbfile}/log/error.log")
    parser_serve.add_argument('--host', nargs="?", type=str, help='The mask to host to', default='127.0.0.1')
    parser_serve.add_argument('--no-addr-reuse', action='store_true')
    parser_serve.add_argument('--debug', action='store_true')
    parser_serve.add_argument('-s', '--scoring_time', nargs="?", type=int, help='The number of seconds between re-scoring all the bots.', default=60)
    parser_serve.set_defaults(func=serve)

    args = parser.parse_args()
    args.func(args)



import random
import sys
import traceback
from abc import ABC, abstractmethod

from littlepython import Compiler, ExecutionCountExceededException
from littlepython.error import LittlePythonBaseExcetion

from CYLGame.Utils import int2base


# A template class for the Prog for the Player.
class Prog(object):
    def __init__(self):
        # The `options`, `token` and `name` attributes are reserved for the GameRunner
        self.options = {}
        self.token = None
        self.name = None

    def run(self, state=None, max_op_count=-1, random=None):
        pass


class LittlePythonProg(Prog):
    def __init__(self, prog, name="Computer"):
        super().__init__()
        self.prog = Compiler().compile(prog)
        self.name = name

    def run(self, *args, **kargs):
        return self.prog.run(*args, **kargs)


class UserProg(Prog):
    def __init__(self):
        super().__init__()
        self.name = "You"
        self.key = None

    def run(self, *args, **kwargs):
        return {"move": ord(self.key)}


class Player(ABC):
    def __init__(self, prog):
        self.prog = prog
        self.prev_vars = {}
        self.debugging = prog.options.get("debug", False)
        self.debug_vars = []  # TODO: Maybe make this into a class.
        self.running = True

    def run_turn(self, random, max_ops=1000000):
        if self.running:
            try:
                nxt_state = self.prog.run(self.get_state(), max_op_count=max_ops, random=random)
                self.update_state(dict(nxt_state))
                self.prev_vars = nxt_state
            except LittlePythonBaseExcetion as e:
                print("Player class raised: {}!".format(e.__class__.__name__))
                self.running = False
                self.prev_vars = {}

    @abstractmethod
    def get_state(self):
        pass

    @abstractmethod
    def update_state(self, new_state):
        pass


class DefaultGridPlayer(Player):
    """Make sure that all bot_vars are updated in game.do_turn"""
    def __init__(self, prog, bot_consts):
        """
        Args:
             prog(Prog):
             bot_consts(ConstMapping):
        """
        super(DefaultGridPlayer, self).__init__(prog)
        self.bot_consts = bot_consts
        self.bot_vars = {}
        self.move = None

    def get_state(self):
        state = dict(self.prev_vars)
        state.update(self.bot_consts)
        state.update(self.bot_vars)
        return state

    def update_state(self, state):
        # remove consts
        for key in self.bot_consts.names:
            state.pop(key, None)

        self.move = chr(state.get("move", ord("Q")))

        if self.debugging:
            human_vars = {}
            for name, val in state.items():
                if isinstance(val, int) and val in self.bot_consts:
                    human_vars[name] = self.bot_consts[val] + " (" + str(val) + ")"
                elif str(val) == str(True):
                    human_vars[name] = 1
                elif str(val) == str(False):
                    human_vars[name] = 0
                else:
                    human_vars[name] = val
            self.debug_vars += [human_vars]


class Room(object):
    def __init__(self, bots=None, seed=None):
        self.bots = bots
        if bots is None:
            self.bots = []

        self.seed = seed
        if seed is None:
            self.seed = random.randint(0, sys.maxsize)

        self.debug_vars = {}
        self.screen_cap = None

    def save(self, gamedb):
        """The method saves the game data to a new game. Note: this room must be run before calling this function.

        Args:
            gamedb (GameDB): The current game database.

        Returns: The gtoken of the newly created game.

        """
        if self.screen_cap is None:
            raise Exception("You must run this room before trying to save it.")
        game_data = {"screen": self.screen_cap,
                     "seed": int2base(self.seed, 36)}
        player_data = {}
        for player in self.bots:
            if hasattr(player, "token") and player.token is not None:
                player_data[player.token] = self.debug_vars[player]
        return gamedb.add_new_game(game_data, per_player_data=player_data)

    @property
    def rand_seeded(self):
        """A randomly seeded room with the same bots."""
        return Room(self.bots)

    def __str__(self):
        return "<Room with '{}'>".format("', '".join(map(lambda x: x.name, self.bots)))
